

/* assemble.h */


/* Opcode argument types: */
#define REG          1  /* register index */
#define INDIRECT     2  /* indirect register {a} */
#define CON1         3  /* 1 byte constant */
#define CON2         4  /* 2 byte constant */
#define RELATIVE     5  /* relative offset */

#define CON          6  /* generic constant (used in expressions) */


/* Arithmetic expression types: */
#define INTEGER_EXPR    1
#define LABEL_EXPR      2
#define ADD_EXPR        3
#define SUBTRACT_EXPR   4
#define MULTIPLY_EXPR   5
#define DIVIDE_EXPR     6
#define MODULUS_EXPR    7
#define AND_EXPR        8
#define OR_EXPR         9
#define XOR_EXPR        10
#define UMINUS_EXPR     11



struct tagExpr
{
  int type; /* one of the arithmetic expression types */
  union
  {
    long integer;     /* INTEGER_EXPR */
    char * label;    /* LABEL_EXPR */
    struct {         /* All others (dyadic operators) */
      struct tagExpr * left;  /* for monadics */
      struct tagExpr * right;
    } operands;
  } value;
};
typedef struct tagExpr Expr;



struct tagArgument
{
  int type;  /* one of the opcode argument types */
  Expr * expr; /* value of argument .. */
  struct tagArgument * next;  /* linked list */
};
typedef struct tagArgument Argument;



struct tagOpcode
{
  char * name;    /* mnemonic name */
  int opcode;     /* hex opcode */
  int args;       /* number of args */
  int * argtypes; /* types of args */
  boolean revargs;  /* reverse first 2 args? */
  struct tagOpcode * next;  /* next in linked list */
};
typedef struct tagOpcode Opcode;



struct tagInstruction
{
  Opcode * opcode;  /* pointer into static opcode table */
  int lineNumber;   /* line number this expression occured in */
  Argument * args;  /* argument list */
  struct tagInstruction * next;
};
typedef struct tagInstruction Instruction;



/* Label list structure */
struct tagLabel
{
  char * name;
  boolean resolved;  /* Resolved yet? */
  Expr * value;      /* EQU value or program counter. */
  struct tagLabel * next;
};
typedef struct tagLabel Label;


