

/* main.c */

#include "main.h"
#include <stdarg.h>


int main(int argc, char * argv[])
{
  FILE * fileptr;
  extern FILE * yyin;

  if(argc != 3)
  {
    printf("syntax: %s <input file> <output file>\n", argv[0]);
    return (1);
  }

  initializeOpcodes();

  /* Try to open the input file. */
  fileptr = fopen(argv[1], "rb");
  if(fileptr == NIL)
  {
    printf("Cannot read file \"%s\".\n", argv[1]);
    return (1);
  }
  yyin = fileptr;

  yyparse();

  if(yyin != stdin && yyin != stdout && yyin != stderr)
    close(yyin);

  /* Open output file. */
  fileptr = fopen(argv[2], "wb");
  if(fileptr == NIL)
  {
    printf("Cannot write file \"%s\".\n", argv[2]);
    return (1);
  }
  generateOutput(fileptr);
  close(fileptr);

  return (0);
}


void error(char * f, ...)
{
  va_list list;
  printf("Line %d: ", lineNumber);
  va_start(list, f);
  vfprintf(stdout, f, list);
  va_end(list);
  printf("\n");
  exit(1);
}


void errorAt(int line, char * f, ...)
{
  va_list list;
  printf("Line %d: ", line);
  va_start(list, f);
  vfprintf(stdout, f, list);
  va_end(list);
  printf("\n");
  exit(1);
}
