

/* opcode.c */

#include "main.h"




/* The main opcode-list. */
Opcode * opcodes = NIL;


/* Opcode-adding functions: */
static void addop0(char * name, int opcode);
static void addop1(char * name, int opcode,
		   int arg1);
static void addop2(char * name, int opcode,
		   int arg1, int arg2);
static void addop3(char * name, int opcode,
		   int arg1, int arg2, int arg3);
static void addrev3(char * name, int opcode,
		    int arg1, int arg2, int arg3);
static Opcode * make_opcode(char * name, int opcode, int args);


/* Add all the opcodes to the table. */
void initializeOpcodes(void)
{
  /* Byte arithmetic: */
  addop2("mov", 0x20, REG, REG);
  addop2("add", 0x21, REG, REG);
  addop2("sub", 0x22, REG, REG);
  addop2("mul", 0x23, REG, REG);
  addop2("div", 0x24, REG, REG);
  addop2("or",  0x25, REG, REG);
  addop2("and", 0x26, REG, REG);
  addop2("xor", 0x27, REG, REG);
  addop2("mod", 0xB5, REG, REG);
  addop2("mov", 0x28, CON1, REG);
  addop2("add", 0x29, CON1, REG);
  addop2("sub", 0x2A, CON1, REG);
  addop2("mul", 0x2B, CON1, REG);
  addop2("div", 0x2C, CON1, REG);
  addop2("or",  0x2D, CON1, REG);
  addop2("and", 0x2E, CON1, REG);
  addop2("xor", 0x2F, CON1, REG);
  addop2("mod", 0xBD, CON1, REG);

  /* Word arithmetic: */
  addop2("mov2", 0x30, REG, REG);
  addop2("add2", 0x31, REG, REG);
  addop2("sub2", 0x32, REG, REG);
  addop2("mul2", 0x33, REG, REG);
  addop2("div2", 0x34, REG, REG);
  addop2("or2",  0x35, REG, REG);
  addop2("and2", 0x36, REG, REG);
  addop2("xor2", 0x37, REG, REG);
  addop2("mod2", 0xB6, REG, REG);
  addop2("mov2", 0x38, CON2, REG);
  addop2("add2", 0x39, CON2, REG);
  addop2("sub2", 0x3A, CON2, REG);
  addop2("mul2", 0x3B, CON2, REG);
  addop2("div2", 0x3C, CON2, REG);
  addop2("or2",  0x3D, CON2, REG);
  addop2("and2", 0x3E, CON2, REG);
  addop2("xor2", 0x3F, CON2, REG);
  addop2("mod2", 0xBE, CON2, REG);

  /* Constants, random numbers & indirect register access: */
  addop2("con", 0xC0, REG, REG);
  addop2("con2", 0xC1, REG, REG);
  addop1("rand", 0xC2, REG);
  addop2("mov", 0xC3, INDIRECT, REG);
  addop2("mov2", 0xC4, INDIRECT, REG);
  addop2("mov", 0xC5, REG, INDIRECT);
  addop2("mov2", 0xC6, REG, INDIRECT);

  /* Program flow: */
  addop1("jump", 0x00, CON2);
  addop1("call", 0x01, CON2);
  addop1("bra", 0x02, RELATIVE);
  addop1("bcs", 0x03, RELATIVE);
  addop1("bcc", 0x04, RELATIVE);
  addop1("bxs", 0x05, RELATIVE);
  addop1("bxc", 0x06, RELATIVE);
  addop0("ret", 0x07);

  /* Test & jump: */
  addop3("beq", 0x10, REG, REG, RELATIVE);
  addop3("bne", 0x11, REG, REG, RELATIVE);
  addop3("bgt", 0x12, REG, REG, RELATIVE);
  addop3("blt", 0x13, REG, REG, RELATIVE);
  addop3("bge", 0x14, REG, REG, RELATIVE);
  addop3("ble", 0x15, REG, REG, RELATIVE);
  addop3("bbs", 0x16, REG, REG, RELATIVE);
  addop3("bbc", 0x17, REG, REG, RELATIVE);
  addop3("beq", 0x18, CON1, REG, RELATIVE);
  addop3("bne", 0x19, CON1, REG, RELATIVE);
  addop3("bgt", 0x1A, CON1, REG, RELATIVE);
  addop3("blt", 0x1B, CON1, REG, RELATIVE);
  addop3("bge", 0x1C, CON1, REG, RELATIVE);
  addop3("ble", 0x1D, CON1, REG, RELATIVE);
  addop3("bbs", 0x1E, CON1, REG, RELATIVE);
  addop3("bbc", 0x1F, CON1, REG, RELATIVE);

  /* reversed logic tests */
  addrev3("beq", 0x18, REG, CON1, RELATIVE);
  addrev3("bne", 0x19, REG, CON1, RELATIVE);
  addrev3("bgt", 0x1B, REG, CON1, RELATIVE);
  addrev3("blt", 0x1A, REG, CON1, RELATIVE);
  addrev3("bge", 0x1D, REG, CON1, RELATIVE);
  addrev3("ble", 0x1C, REG, CON1, RELATIVE);
  addrev3("bbs", 0x1E, REG, CON1, RELATIVE);
  addrev3("bbc", 0x1F, REG, CON1, RELATIVE);


  /* Double-byte test & jump: */
  addop3("beq2", 0x40, REG, REG, RELATIVE);
  addop3("bne2", 0x41, REG, REG, RELATIVE);
  addop3("bgt2", 0x42, REG, REG, RELATIVE);
  addop3("blt2", 0x43, REG, REG, RELATIVE);
  addop3("bge2", 0x44, REG, REG, RELATIVE);
  addop3("ble2", 0x45, REG, REG, RELATIVE);
  addop3("bbs2", 0x46, REG, REG, RELATIVE);
  addop3("bbc2", 0x47, REG, REG, RELATIVE);
  addop3("beq2", 0x48, CON2, REG, RELATIVE);
  addop3("bne2", 0x49, CON2, REG, RELATIVE);
  addop3("bgt2", 0x4A, CON2, REG, RELATIVE);
  addop3("blt2", 0x4B, CON2, REG, RELATIVE);
  addop3("bge2", 0x4C, CON2, REG, RELATIVE);
  addop3("ble2", 0x4D, CON2, REG, RELATIVE);
  addop3("bbs2", 0x4E, CON2, REG, RELATIVE);
  addop3("bbc2", 0x4F, CON2, REG, RELATIVE);

  /* reversed logic tests */
  addrev3("beq2", 0x48, REG, CON2, RELATIVE);
  addrev3("bne2", 0x49, REG, CON2, RELATIVE);
  addrev3("bgt2", 0x4B, REG, CON2, RELATIVE);
  addrev3("blt2", 0x4A, REG, CON2, RELATIVE);
  addrev3("bge2", 0x4D, REG, CON2, RELATIVE);
  addrev3("ble2", 0x4C, REG, CON2, RELATIVE);
  addrev3("bbs2", 0x4E, REG, CON2, RELATIVE);
  addrev3("bbc2", 0x4F, REG, CON2, RELATIVE);


  /* Keyboard: */
  addop2("bkp", 0x60, CON1, RELATIVE);
  addop2("bkr", 0x61, CON1, RELATIVE);

  /* Timer: */
  addop1("tim", 0x70, REG);
  addop2("btg", 0x71, REG, RELATIVE);
  addop2("btl", 0x72, REG, RELATIVE);

  /* Screen handling: */
  addop0("cls", 0xA0);
  addop0("scr", 0xA1);
  addop0("scl", 0xA2);
  addop3("sor", 0x50, REG, REG, REG);
  addop3("sxor", 0x51, REG, REG, REG);
  addop3("snot", 0x52, REG, REG, REG);
  addop3("snum", 0x53, REG, REG, REG);
  addop3("sdig", 0x54, REG, REG, REG);
  addop1("scu", 0x55, REG);
  addop1("scd", 0x56, REG);
  addop3("sor", 0x58, REG, REG, CON2);
  addop3("sxor", 0x59, REG, REG, CON2);
  addop3("snot", 0x5A, REG, REG, CON2);
  addop3("snum", 0x5B, CON1, CON1, REG);
  addop3("sdig", 0x5C, CON1, CON1, REG);
  addop1("scu", 0x5D, CON1);
  addop1("scd", 0x5E, CON1);
  addop2("pxs", 0xB0, REG, REG);
  addop2("pxc", 0xB1, REG, REG);
  addop2("pxt", 0xB2, REG, REG);
  addop3("bps", 0xB3, REG, REG, RELATIVE);
  addop3("bpc", 0xB4, REG, REG, RELATIVE);
  addop2("pxs", 0xB8, CON1, CON1);
  addop2("pxt", 0xBA, CON1, CON1);
  addop3("bps", 0xBB, CON1, CON1, RELATIVE);
  addop3("bpc", 0xBC, CON1, CON1, RELATIVE);

  /* Sound: */
  addop1("snd0", 0x80, CON1);
  addop1("snd1", 0x81, CON1);
  addop1("snd2", 0x82, CON1);
  addop1("snd3", 0x83, CON1);
  addop1("snd4", 0x84, CON1);
  addop1("snd5", 0x85, CON1);
  addop1("snd6", 0x86, CON1);
  addop1("snd7", 0x87, CON1);
  addop1("snd8", 0x88, CON1);
  addop1("snd9", 0x89, CON1);
  addop1("snda", 0x8A, CON1);
  addop1("sndb", 0x8B, CON1);
  addop1("sndc", 0x8C, CON1);
  addop1("sndd", 0x8D, CON1);
  addop1("snde", 0x8E, CON1);
  addop1("sndf", 0x8F, CON1);

  /* Interpreter control: */
  addop0("exit", 0x90);


  /* Special opcodes: */
  addop1("db", -1, CON1);  /* byte data */
  addop1("dw", -2, CON2);  /* word data */
  addop1("ds", -3, CON2);  /* sprite data */
}


/* Locate an opcode with the given name and arguments.
   Returns NIL if the opcode/argument combo does not exist. */
Opcode * matchOpcode(char * name, Argument * args)
{
  int numArgs;
  Argument * argptr;
  Opcode * opptr;
  int * intptr;

  /* Count number of argumnents. */
  for(numArgs = 0, argptr = args; argptr != NIL;
      argptr = argptr->next, numArgs++)
    ;

  /* Loop through each opcode. */
  for(opptr = opcodes; opptr != NIL; opptr = opptr->next)
  {
    if(STRCASECMP(opptr->name, name) == 0)
    {
      /* Name matches ... do the argument types? */
      /* NOTE: only superficial matching is done
	  at this phase, since we do not know
	  most label values. */
      if(opptr->args != numArgs)
	continue;  /* bad # of args */

      /* Loop through each argument type to make sure they
	 all match... */
      for(argptr = args, intptr = opptr->argtypes;
	  argptr != NIL; argptr = argptr->next, intptr++)
      {
	boolean matched;
	matched = FALSE;
	switch(argptr->type)
	{
	case REG:
  	  if(*intptr == REG)
	    matched = TRUE;
	  break;
	case INDIRECT:
	  if(*intptr == INDIRECT)
	    matched = TRUE;
	  break;
	case CON:
	  if(*intptr == CON1 || *intptr == CON2 ||
	     *intptr == RELATIVE)
	    matched = TRUE;
	  break;
	default:
	  error("internal error: bad preliminary argtype %d",
		argptr->type);
	}
	if(matched == FALSE)
	  break;
      }
      if(argptr != NIL)
	continue;  /* not successful */

      /* Otherwise, this instruction has a good
	 chance of being ok... */
      return (opptr);
    }
  }
  return (NIL);
}


/* Get the length in bytes of an opcode. */
int opcodeLength(Opcode * op)
{
  int len;
  int index;

  if(op->opcode < 0)
    len = 0;
  else len = 1;
  for(index = 0; index < op->args; index++)
  {
    if(op->argtypes[index] == CON2)
      len += 2;  /* Large constant = 2 bytes */
    else
      len += 1;  /* All others = 1 byte */
  }
  return (len);
}



static void addop0(char * name, int opcode)
{
  make_opcode(name, opcode, 0);
}


static void addop1(char * name, int opcode,
		   int arg1)
{
  Opcode * op = make_opcode(name, opcode, 1);
  op->argtypes[0] = arg1;
}
		   

static void addop2(char * name, int opcode,
		   int arg1, int arg2)
{
  Opcode * op = make_opcode(name, opcode, 2);
  op->argtypes[0] = arg1;
  op->argtypes[1] = arg2;
}


static void addop3(char * name, int opcode,
		   int arg1, int arg2, int arg3)
{
  Opcode * op = make_opcode(name, opcode, 3);
  op->argtypes[0] = arg1;
  op->argtypes[1] = arg2;
  op->argtypes[2] = arg3;
}

static void addrev3(char * name, int opcode,
		    int arg1, int arg2, int arg3)
{
  Opcode * op = make_opcode(name, opcode, 3);
  op->argtypes[0] = arg1;
  op->argtypes[1] = arg2;
  op->argtypes[2] = arg3;
  op->revargs = TRUE;
}



static Opcode * make_opcode(char * name, int opcode, int args)
{
  Opcode * op;
  op = allocate(sizeof(Opcode));
  op->name = duplicateString(name);
  op->opcode = opcode;
  op->args = args;
  op->revargs = FALSE;
  if(args > 0)
    op->argtypes = (int *)allocate(sizeof(int) * args);
  else op->argtypes = NIL;
  op->next = opcodes;
  opcodes = op;
  return (op);
}
