

%{

#include "main.h"


/* Static functions: */

static Expr * makeIntegerExpr(long value);
static Expr * makeLabelExpr(char * label);
static Expr * makeDyadicExpr(int op, Expr * left, Expr * right);
static Expr * makeMonadicExpr(int op, Expr * operand);

static Argument * makeArgument(int type, Expr * expr);
static void defineLabel(char * labelName, Expr * value);

%}


%union {
  long integer;
  char * string;
  Expr * expr;
  Argument * argument;
}

 
%token EQU NEWLINE
%token <integer> INTEGER
%token <string> LABEL_DEFINITION IDENT

%type <expr> constant expr
%type <argument> argument argumentList


%left ','
%left '|' '^' '&'
%left '+' '-'
%left '*' '/' '%'



%%


start:
	/* empty */
|	start inputUnit
;


inputUnit:
	labelDefinition
|	label
|	opcode
|	labelAndOpcode
|	emptyLine
;


label:
	LABEL_DEFINITION NEWLINE
	{ defineLabel($1, makeIntegerExpr(pc)); }
;

labelDefinition:
	LABEL_DEFINITION EQU expr NEWLINE
	{ defineLabel($1, $3); }
;

opcode:
	IDENT argumentList NEWLINE
	{ Opcode * op;
	  op = matchOpcode($1, $2);
          if(op == NIL)
            error("opcode \"%s\" is invalid", $1);
          gentemp(op, $2); }
;

labelAndOpcode:
	LABEL_DEFINITION IDENT argumentList NEWLINE
	{ Opcode * op;
          defineLabel($1, makeIntegerExpr(pc));
          op = matchOpcode($2, $3);
	  if(op == NIL)
            error("opcode \"%s\" is invalid", $2);
          gentemp(op, $3); }
;

emptyLine:
	NEWLINE
;


argumentList:
	/* empty */		  { $$ = NIL; }
|	argument                  { $$ = $1; }
|	argument ',' argumentList
	{ $1->next = $3;
	  $$ = $1; }
;

argument:
	expr		{ $$ = makeArgument(CON, $1); }
|	'[' expr ']'	{ $$ = makeArgument(REG, $2); }
|	'{' expr '}'    { $$ = makeArgument(INDIRECT, $2); }
;


expr:
	constant
|	expr '+' expr
	{ $$ = makeDyadicExpr(ADD_EXPR, $1, $3); }
|	expr '-' expr
	{ $$ = makeDyadicExpr(SUBTRACT_EXPR, $1, $3); }
|	expr '*' expr
	{ $$ = makeDyadicExpr(MULTIPLY_EXPR, $1, $3); }
|	expr '/' expr
	{ $$ = makeDyadicExpr(DIVIDE_EXPR, $1, $3); }
|	expr '%' expr
	{ $$ = makeDyadicExpr(MODULUS_EXPR, $1, $3); }
|	expr '&' expr
	{ $$ = makeDyadicExpr(AND_EXPR, $1, $3); }
|	expr '|' expr
	{ $$ = makeDyadicExpr(OR_EXPR, $1, $3); }
|	expr '^' expr
	{ $$ = makeDyadicExpr(XOR_EXPR, $1, $3); }
|	'(' expr ')'
	{ $$ = $2; }
;


constant:
	INTEGER
        { if($1 < 0)
	    error("negative constants (%d) are not allowed", $1);
          if($1 > 0xFFFFl)
            error("constant %d too large (%d max)", $1, 0xFFFFl);
          $$ = makeIntegerExpr($1); }
|	IDENT
	{ $$ = makeLabelExpr($1);
	  free($1); }
;



%%


static Expr * makeIntegerExpr(long value)
{
  Expr * expr = (Expr *)allocate(sizeof(Expr));
  expr->type = INTEGER_EXPR;
  expr->value.integer = value;
  return (expr);
}


static Expr * makeLabelExpr(char * label)
{
  Expr * expr = (Expr *)allocate(sizeof(Expr));
  expr->type = LABEL_EXPR;
  expr->value.label = duplicateString(label);
  return (expr);
}


static Expr * makeDyadicExpr(int op, Expr * left, Expr * right)
{
  Expr * expr = (Expr *)allocate(sizeof(Expr));
  expr->type = op;
  expr->value.operands.left = left;
  expr->value.operands.right = right;
  return (expr);
}


static Expr * makeMonadicExpr(int op, Expr * operand)
{
  Expr * expr = (Expr *)allocate(sizeof(Expr));
  expr->type = op;
  expr->value.operands.left = operand;
  expr->value.operands.right = NIL;
  return (expr);
}


static Argument * makeArgument(int type, Expr * expr)
{
  Argument * argument = (Argument *)allocate(sizeof(Argument));
  argument->type = type;
  argument->expr = expr;
  argument->next = NIL;
  return (argument);
}


/* Define a label with the given value. */
static void defineLabel(char * labelName, Expr * value)
{
  Label * label;
  label = findLabel(labelName);
  if(label == NIL)
    label = addLabel(labelName);
  if(label->resolved == TRUE)
    error("cannot redefine label \"%s\"", labelName);     
  resolveLabel(label, value);
}
