

%{

/* scan.l */

#include "main.h"
#include "ytab.h"

int lineNumber = 1;


static long parseBinary(char * str);
static long parseHex(char * str);


%}


IDENT	[a-zA-Z_][a-zA-Z_0-9]*


%%


[0-9]+		{ yylval.integer = atol(yytext);
		  return (INTEGER); }

[0-1]+b         { yylval.integer = parseBinary(yytext);
                  return (INTEGER); }

"#"[0-1]+b	{ yylval.integer = parseBinary(&(yytext[1]));
                  return (INTEGER); }

[0-9a-fA-F]+h   { yylval.integer = parseHex(yytext);
                  return (INTEGER); }

"#"[0-9a-fA-F]+h { yylval.integer = parseHex(&(yytext[1]));
                   return (INTEGER); }

[eE][qQ][uU]	{ return (EQU); }

^{IDENT}	{ yylval.string = duplicateString(yytext);
		  return (LABEL_DEFINITION); }

{IDENT}		{ yylval.string = duplicateString(yytext);
		  return (IDENT); }


[ \t\r\f]	/* eat white space */

\n		{ lineNumber++; return (NEWLINE); }

";".*		/* eat comments */

.		{ return (yytext[0]); }


%%


int yywrap()
{
  return (1); 
}


int yyerror(char * s)
{
  error("%s near \"%s\"", s, yytext);
  return (0);
}



/* Parse a binary number.  The number is guaranteed
   to end in 'b'. */
static long parseBinary(char * str)
{
  long value = 0L;
  while(*str != 'b')
  {
    value <<= 1;
    if(*str == '1')
      value |= 1;
    str++;
  }
  return (value);
}


/* Parse a hexidecimal number.  The number is guaranteed
   to end in 'h'. */
static long parseHex(char * str)
{
  long value = 0L;
  int digitvalue;

  while(*str != 'h')
  {
    value <<= 4;
    if(*str >= '0' && *str <= '9')
      digitvalue = (*str) - '0';
    else if(*str >= 'a' && *str <= 'f')
      digitvalue = ((*str) - 'a') + 10;
    else if(*str >= 'A' && *str <= 'F')
      digitvalue = ((*str) - 'A') + 10;
    value |= (digitvalue & 15);
    str++;
  }
  return (value);
}
